# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dataclasses
from typing import Any, Generic, Literal, TypeVar, overload

import numpy as np
import numpy.typing as npt
import optype.numpy as onp

from ._spec_array_object import array
from ._typing import Dtype, Shape

@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.ToDType[_SCT],
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[_SCT]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: None,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Any]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: type[bool] | Literal[
        "?",
        "=?",
        "<?",
        ">?",
        "bool",
        "bool_",
        "bool8",
    ],
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyInt8DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.int8]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyInt16DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.int16]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyInt32DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.int32]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyInt64DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.int64]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyUInt8DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.uint8]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyUInt16DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.uint16]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyUInt32DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.uint32]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyUInt64DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.uint64]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyFloat32DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.float32]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyFloat64DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.float64]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyComplex64DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.complex64]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: onp.AnyComplex128DType,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT, Dtype[np.complex128]]: ...
@overload
def astype(
    x: array[_ShapeT],
    dtype: type[Any] | str,
    /,
    *,
    copy: bool = ...,
) -> array[_ShapeT]: ...

def can_cast(from_: object, to: npt.DTypeLike, /) -> bool: ...

@dataclasses.dataclass
class finfo_object(Generic[_FloatingT_co]):  # noqa: N801
    bits: int
    eps: _FloatingT_co
    max: _FloatingT_co
    min: _FloatingT_co
    smallest_normal: _FloatingT_co
    dtype: np.dtype[_FloatingT_co]

@overload
def finfo(
    type: onp.ToDType[_FloatingT_co],
) -> finfo_object[_FloatingT_co]: ...
@overload
def finfo(
    type: onp.ToDType[np.complexfloating[_BitT, _BitT]],
) -> finfo_object[np.floating[_BitT]]: ...
@overload
def finfo(
    type: onp.AnyFloat32DType | onp.AnyComplex64DType,
) -> finfo_object[np.float32]: ...
@overload
def finfo(
    type: onp.AnyFloat64DType | onp.AnyComplex128DType,
) -> finfo_object[np.float64]: ...

@dataclasses.dataclass
class iinfo_object(Generic[_IntegerT_co]):  # noqa: N801
    bits: int
    max: int
    min: int
    dtype: np.dtype[_IntegerT_co]

@overload
def iinfo(type: onp.ToDType[_IntegerT_co]) -> iinfo_object[_IntegerT_co]: ...
@overload
def iinfo(type: onp.AnyInt8DType) -> iinfo_object[np.int8]: ...
@overload
def iinfo(type: onp.AnyInt16DType) -> iinfo_object[np.int16]: ...
@overload
def iinfo(type: onp.AnyInt32DType) -> iinfo_object[np.int32]: ...
@overload
def iinfo(type: onp.AnyInt64DType) -> iinfo_object[np.int64]: ...
@overload
def iinfo(type: onp.AnyUInt8DType) -> iinfo_object[np.uint8]: ...
@overload
def iinfo(type: onp.AnyUInt16DType) -> iinfo_object[np.uint16]: ...
@overload
def iinfo(type: onp.AnyUInt32DType) -> iinfo_object[np.uint32]: ...
@overload
def iinfo(type: onp.AnyUInt64DType) -> iinfo_object[np.uint64]: ...

def isdtype(
    dtype: Dtype,
    kind: Dtype | str | tuple[Dtype | str, ...],
) -> bool: ...

def result_type(*arrays_and_dtypes: object) -> np.dtype[Any]: ...

_BitT = TypeVar("_BitT", bound=npt.NBitBase)
_FloatingT_co = TypeVar(
    "_FloatingT_co",
    bound=np.floating[Any],
    covariant=True,
)
_IntegerT_co = TypeVar("_IntegerT_co", bound=np.integer[Any], covariant=True)
_SCT = TypeVar("_SCT", bound=np.bool_ | np.number[Any])
_ShapeT = TypeVar("_ShapeT", bound=Shape)



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import enum
from collections.abc import Sequence
from typing import Any, Literal, SupportsIndex, TypeAlias, TypeVar, overload

import numpy as np
import optype as op
import optype.numpy as onp
from optype.dlpack import CanDLPackDevice
from typing_extensions import Unpack

from ._spec_array_object import array
from ._typing import Device, Dtype, Shape

@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int], Dtype[_SCT]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.bool_]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int8]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int16]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int32]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int64]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint8]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint16]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint32]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint64]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float32]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float64]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex64]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex128]]: ...
@overload
def arange(
    start: object,
    /,
    stop: object = ...,
    step: object = ...,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int]]: ...

@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[_SCT]]: ...
@overload
def asarray(
    obj: array[_ShapeT, _DTypeT],
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, _DTypeT]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: _Bool,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.int8]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.int16]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.int32]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.int64]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.uint8]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.uint16]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.uint32]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.uint64]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.float32]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.float64]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.complex64]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT, Dtype[np.complex128]]: ...
@overload
def asarray(
    obj: array[_ShapeT],
    dtype: type[Any] | str,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_ShapeT]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[_SCT]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _DTypeT],
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, _DTypeT]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: _Bool,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.bool_]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.int8]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.int16]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.int32]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.int64]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint8]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint16]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint32]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint64]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.float32]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.float64]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex64]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex128]]: ...
@overload
def asarray(
    obj: np.ndarray[_RegularShapeT, _NumPyDType],
    dtype: type[Any] | str,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[_RegularShapeT]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[_SCT]]: ...
@overload
def asarray(
    obj: _NumberT,
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[_NumberT]]: ...
@overload
def asarray(
    obj: bool,
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.bool_]]: ...
@overload
def asarray(
    obj: op.JustInt,
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.int64]]: ...
@overload
def asarray(
    obj: op.JustFloat,
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.float64]]: ...
@overload
def asarray(
    obj: op.JustComplex,
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.complex128]]: ...
@overload
def asarray(
    obj: object,
    dtype: None = ...,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: _Bool,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.bool_]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.int8]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.int16]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.int32]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.int64]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.uint8]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.uint16]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.uint32]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.uint64]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.float32]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.float64]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.complex64]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()], Dtype[np.complex128]]: ...
@overload
def asarray(
    obj: complex | np.number[Any] | np.timedelta64,
    dtype: type[Any] | str,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[tuple[()]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[_SCT]]: ...
@overload
def asarray(
    obj: object,
    dtype: _Bool,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.bool_]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.int8]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.int16]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.int32]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.int64]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.uint8]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.uint16]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.uint32]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.uint64]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.float32]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.float64]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.complex64]]: ...
@overload
def asarray(
    obj: object,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array[Any, Dtype[np.complex128]]: ...
@overload
def asarray(
    obj: object,
    dtype: type[Any] | str,
    device: Device | None = ...,
    copy: bool | None = ...,
) -> array: ...

@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[_SCT]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int], Dtype[_SCT]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[_SCT]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float64]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float64]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float64]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.bool_]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.bool_]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.bool_]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int8]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int8]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int8]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int16]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int16]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int16]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int32]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int32]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int32]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int64]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int64]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int64]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint8]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint8]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint8]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint16]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint16]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint16]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint32]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint32]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint32]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint64]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint64]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint64]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float32]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float32]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float32]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex64]]: ...
@overload
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex64]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex64]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex128]]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex128]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex128]]: ...
@overload
def empty(
    shape: _RegularShapeT,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_RegularShapeT]: ...
@overload
def empty(
    shape: SupportsIndex,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int]]: ...
@overload
def empty(
    shape: Sequence[SupportsIndex],
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int, ...]]: ...

@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[_SCT]]: ...
@overload
def empty_like(
    x: array[_ShapeT, _DTypeT],
    /,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_ShapeT, _DTypeT]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int8]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int16]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int32]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int64]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint8]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint16]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint32]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint64]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float32]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float64]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex64]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex128]]: ...
@overload
def empty_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_ShapeT]: ...

@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[_SCT]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.float64]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.bool_]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.int8]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.int16]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.int32]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.int64]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.uint8]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.uint16]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.uint32]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.uint64]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.float32]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.complex64]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int, int], Dtype[np.complex128]]: ...
@overload
def eye(
    n_rows: int,
    n_cols: int | None = ...,
    /,
    *,
    k: int = ...,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int, int]]: ...

def from_dlpack(
    x: CanDLPackDevice[enum.Enum | int, int],
    /,
) -> array: ...

@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[_SCT]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int], Dtype[_SCT]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[_SCT]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: _SCT,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[_SCT]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: _SCT,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[_SCT]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: _SCT,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[_SCT]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: bool,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.bool_]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: bool,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.bool_]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: bool,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.bool_]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: op.JustInt,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int64]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: op.JustInt,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int64]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: op.JustInt,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int64]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: op.JustFloat,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float64]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: op.JustFloat,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float64]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: op.JustFloat,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float64]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: op.JustComplex,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex128]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: op.JustComplex,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex128]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: op.JustComplex,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex128]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.bool_]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.bool_]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.bool_]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int8]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int8]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int8]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int16]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int16]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int16]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int32]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int32]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int32]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int64]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int64]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int64]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint8]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint8]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint8]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint16]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint16]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint16]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint32]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint32]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint32]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint64]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint64]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint64]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float32]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float32]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float32]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float64]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float64]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float64]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex64]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex64]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex64]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex128]]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex128]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex128]]: ...
@overload
def full(
    shape: _RegularShapeT,
    fill_value: object,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_RegularShapeT]: ...
@overload
def full(
    shape: SupportsIndex,
    fill_value: object,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int]]: ...
@overload
def full(
    shape: Sequence[SupportsIndex],
    fill_value: object,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int, ...]]: ...

@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[_SCT]]: ...
@overload
def full_like(
    x: array[_ShapeT, _DTypeT],
    /,
    fill_value: object,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_ShapeT, _DTypeT]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int8]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int16]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int32]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int64]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint8]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint16]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint32]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint64]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float32]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float64]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex64]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex128]]: ...
@overload
def full_like(
    x: array[_ShapeT],
    /,
    fill_value: object,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_ShapeT]: ...

@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[_SCT]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: None = ...,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.inexact[Any]]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: _Bool,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.bool_]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.int8]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.int16]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.int32]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.int64]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.uint8]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.uint16]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.uint32]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.uint64]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.float32]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.float64]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.complex64]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D, Dtype[np.complex128]]: ...
@overload
def linspace(
    start: object,
    stop: object,
    /,
    num: int,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
    endpoint: bool = ...,
) -> array[onp.AtLeast1D]: ...

def meshgrid(
    *arrays: array,
    indexing: Literal["xy", "ij"] = ...,
) -> list[array]: ...

@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[_SCT]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int], Dtype[_SCT]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[_SCT]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float64]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float64]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float64]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.bool_]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.bool_]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.bool_]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int8]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int8]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int8]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int16]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int16]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int16]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int32]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int32]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int32]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int64]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int64]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int64]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint8]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint8]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint8]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint16]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint16]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint16]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint32]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint32]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint32]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint64]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint64]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint64]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float32]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float32]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float32]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex64]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex64]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex64]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex128]]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex128]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex128]]: ...
@overload
def ones(
    shape: _RegularShapeT,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_RegularShapeT]: ...
@overload
def ones(
    shape: SupportsIndex,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int]]: ...
@overload
def ones(
    shape: Sequence[SupportsIndex],
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int, ...]]: ...

@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[_SCT]]: ...
@overload
def ones_like(
    x: array[_ShapeT, _DTypeT],
    /,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_ShapeT, _DTypeT]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int8]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int16]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int32]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int64]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint8]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint16]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint32]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint64]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float32]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float64]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex64]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex128]]: ...
@overload
def ones_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_ShapeT]: ...

def tril(
    x: array[Any, _DTypeT],
    /,
    *,
    k: int = ...,
) -> array[tuple[int, int] | tuple[int, int, int], _DTypeT]: ...

def triu(
    x: array[Any, _DTypeT],
    /,
    *,
    k: int = ...,
) -> array[tuple[int, int] | tuple[int, int, int], _DTypeT]: ...

@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[_SCT]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int], Dtype[_SCT]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[_SCT]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float64]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float64]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyFloat64DType | None = ...,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float64]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.bool_]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.bool_]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.bool_]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int8]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int8]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int8]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int16]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int16]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int16]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int32]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int32]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int32]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.int64]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.int64]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.int64]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint8]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint8]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint8]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint16]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint16]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint16]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint32]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint32]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint32]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.uint64]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.uint64]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.uint64]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.float32]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.float32]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.float32]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex64]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex64]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex64]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_RegularShapeT, Dtype[np.complex128]]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int], Dtype[np.complex128]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[tuple[int, ...], Dtype[np.complex128]]: ...
@overload
def zeros(
    shape: _RegularShapeT,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_RegularShapeT]: ...
@overload
def zeros(
    shape: SupportsIndex,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int]]: ...
@overload
def zeros(
    shape: Sequence[SupportsIndex],
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[tuple[int, ...]]: ...

@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.ToDType[_SCT],
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[_SCT]]: ...
@overload
def zeros_like(
    x: array[_ShapeT, _DTypeT],
    /,
    *,
    dtype: None = ...,
    device: Device | None = ...,
) -> array[_ShapeT, _DTypeT]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: _Bool,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int8]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int16]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int32]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.int64]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt8DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint8]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt16DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint16]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint32]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyUInt64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.uint64]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyFloat32DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float32]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyFloat64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.float64]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyComplex64DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex64]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: onp.AnyComplex128DType,
    device: Device | None = ...,
) -> array[_ShapeT, Dtype[np.complex128]]: ...
@overload
def zeros_like(
    x: array[_ShapeT],
    /,
    *,
    dtype: type[Any] | str,
    device: Device | None = ...,
) -> array[_ShapeT]: ...

_Bool: TypeAlias = type[bool] | Literal[
    "?",
    "=?",
    "<?",
    ">?",
    "bool",
    "bool_",
    "bool8",
]
_DTypeT = TypeVar("_DTypeT", bound=Dtype)
_NumberT = TypeVar("_NumberT", bound=np.number[Any])
_NumPyDType: TypeAlias = np.dtype[
    np.bool_ | np.number[Any] | np.datetime64 | np.timedelta64
]
_RegularShapeT = TypeVar(
    "_RegularShapeT",
    tuple[()],
    tuple[int],
    tuple[int, int],
    tuple[int, int, int],
    tuple[int, int, int, int],
    tuple[int, int, int, int, Unpack[tuple[int, ...]]],
    tuple[int, int, int, Unpack[tuple[int, ...]]],
    tuple[int, int, Unpack[tuple[int, ...]]],
    tuple[int, Unpack[tuple[int, ...]]],
    tuple[int, ...],
)
_SCT = TypeVar("_SCT", bound=np.bool_ | np.number[Any])
_ShapeT = TypeVar("_ShapeT", bound=Shape)



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__all__ = [  # noqa: RUF022
    "__array_api_version__",
    # _spec_array_object
    "array",
    # _spec_constants
    "e",
    "inf",
    "nan",
    "newaxis",
    "pi",
    # _spec_creation_functions
    "arange",
    "asarray",
    "empty",
    "empty_like",
    "eye",
    "from_dlpack",
    "full",
    "full_like",
    "linspace",
    "meshgrid",
    "ones",
    "ones_like",
    "tril",
    "triu",
    "zeros",
    "zeros_like",
    # _spec_data_type_functions
    "astype",
    "can_cast",
    "finfo",
    "iinfo",
    "isdtype",
    "result_type",
    # _spec_elementwise_functions
    "abs",
    "acos",
    "acosh",
    "add",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "bitwise_and",
    "bitwise_left_shift",
    "bitwise_invert",
    "bitwise_or",
    "bitwise_right_shift",
    "bitwise_xor",
    "ceil",
    "conj",
    "cos",
    "cosh",
    "divide",
    "equal",
    "exp",
    "expm1",
    "floor",
    "floor_divide",
    "greater",
    "greater_equal",
    "imag",
    "isfinite",
    "isinf",
    "isnan",
    "less",
    "less_equal",
    "log",
    "log1p",
    "log2",
    "log10",
    "logaddexp",
    "logical_and",
    "logical_not",
    "logical_or",
    "logical_xor",
    "multiply",
    "negative",
    "not_equal",
    "positive",
    "pow",
    "real",
    "remainder",
    "round",
    "sign",
    "sin",
    "sinh",
    "square",
    "sqrt",
    "subtract",
    "tan",
    "tanh",
    "trunc",
    # _spec_indexing_functions
    "take",
    # _spec_linear_algebra_functions
    "matmul",
    "matrix_transpose",
    "tensordot",
    "vecdot",
    # _spec_manipulation_functions
    "broadcast_arrays",
    "broadcast_to",
    "concat",
    "expand_dims",
    "flip",
    "permute_dims",
    "reshape",
    "roll",
    "squeeze",
    "stack",
    # _spec_searching_functions
    "argmax",
    "argmin",
    "nonzero",
    "where",
    # _spec_set_functions
    "unique_all",
    "unique_counts",
    "unique_inverse",
    "unique_values",
    # _spec_sorting_functions
    "argsort",
    "sort",
    # _spec_statistical_functions
    "max",
    "mean",
    "min",
    "prod",
    "std",
    "sum",
    "var",
    # _spec_utility_functions
    "all",
    "any",
]

from typing import Final

from ._spec_array_object import array
from ._spec_constants import (
    e,
    inf,
    nan,
    newaxis,
    pi,
)
from ._spec_creation_functions import (
    arange,
    asarray,
    empty,
    empty_like,
    eye,
    from_dlpack,
    full,
    full_like,
    linspace,
    meshgrid,
    ones,
    ones_like,
    tril,
    triu,
    zeros,
    zeros_like,
)
from ._spec_data_type_functions import (
    astype,
    can_cast,
    finfo,
    iinfo,
    isdtype,
    result_type,
)
from ._spec_elementwise_functions import (
    abs,  # noqa: A004
    acos,
    acosh,
    add,
    asin,
    asinh,
    atan,
    atan2,
    atanh,
    bitwise_and,
    bitwise_invert,
    bitwise_left_shift,
    bitwise_or,
    bitwise_right_shift,
    bitwise_xor,
    ceil,
    conj,
    cos,
    cosh,
    divide,
    equal,
    exp,
    expm1,
    floor,
    floor_divide,
    greater,
    greater_equal,
    imag,
    isfinite,
    isinf,
    isnan,
    less,
    less_equal,
    log,
    log1p,
    log2,
    log10,
    logaddexp,
    logical_and,
    logical_not,
    logical_or,
    logical_xor,
    multiply,
    negative,
    not_equal,
    positive,
    pow,  # noqa: A004
    real,
    remainder,
    round,  # noqa: A004
    sign,
    sin,
    sinh,
    sqrt,
    square,
    subtract,
    tan,
    tanh,
    trunc,
)
from ._spec_indexing_functions import (
    take,
)
from ._spec_linear_algebra_functions import (
    matmul,
    matrix_transpose,
    tensordot,
    vecdot,
)
from ._spec_manipulation_functions import (
    broadcast_arrays,
    broadcast_to,
    concat,
    expand_dims,
    flip,
    permute_dims,
    reshape,
    roll,
    squeeze,
    stack,
)
from ._spec_searching_functions import (
    argmax,
    argmin,
    nonzero,
    where,
)
from ._spec_set_functions import (
    unique_all,
    unique_counts,
    unique_inverse,
    unique_values,
)
from ._spec_sorting_functions import (
    argsort,
    sort,
)
from ._spec_statistical_functions import (
    max,  # noqa: A004
    mean,
    min,  # noqa: A004
    prod,
    std,
    sum,  # noqa: A004
    var,
)
from ._spec_utility_functions import (
    all,  # noqa: A004
    any,  # noqa: A004
)

__array_api_version__: Final[str]



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import enum
from typing import Any, Final, Literal, TypeAlias

import numpy as np
from optype.dlpack import CanDLPackCompat, CanDLPackDevice
from optype.numpy import SequenceND
from typing_extensions import Buffer, CapsuleType, TypeVar, Unpack

_SCT = TypeVar(
    "_SCT",
    bound=np.bool_ | np.number[Any],
    default=np.bool_ | np.number[Any],
)

Device: TypeAlias = Literal["cpu", "cuda"]
Dtype: TypeAlias = np.dtype[_SCT]
NestedSequence: TypeAlias = SequenceND[_T_co]
PyCapsule: TypeAlias = CapsuleType
Shape: TypeAlias = tuple[int, ...] | tuple[int, Unpack[tuple[int | None, ...]]]
SupportsBufferProtocol: TypeAlias = Buffer

class SupportsDLPack(CanDLPackCompat, CanDLPackDevice[enum.Enum, int]): ...

numeric_types: Final[tuple[
    type[np.bool_],
    type[np.int8],
    type[np.int16],
    type[np.int32],
    type[np.int64],
    type[np.uint8],
    type[np.uint16],
    type[np.uint32],
    type[np.uint64],
    type[np.float32],
    type[np.float64],
    type[np.complex64],
    type[np.complex128],
]]

_T_co = TypeVar("_T_co", covariant=True)



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Generic, TypeVar

import numpy as np
from typing_extensions import NamedTuple

from ._spec_array_object import array
from ._typing import Dtype

class unique_all_result(NamedTuple, Generic[_DTypeT_co]):  # noqa: N801
    values: array[tuple[int], _DTypeT_co]
    indices: array[tuple[int], Dtype[np.int64]]
    inverse_indices: array[tuple[int], Dtype[np.int64]]
    counts: array[tuple[int], Dtype[np.int64]]

def unique_all(x: array[Any, _DTypeT], /) -> unique_all_result[_DTypeT]: ...

class unique_counts_result(NamedTuple, Generic[_DTypeT_co]):  # noqa: N801
    values: array[tuple[int], _DTypeT_co]
    counts: array[tuple[int], Dtype[np.int64]]

def unique_counts(x: array[Any, _DTypeT], /) -> unique_counts_result[_DTypeT]:
    ...
class unique_inverse_result(NamedTuple, Generic[_DTypeT_co]):  # noqa: N801
    values: array[tuple[int], _DTypeT_co]
    inverse_indices: array[tuple[int], Dtype[np.int64]]

def unique_inverse(
    x: array[Any, _DTypeT],
    /,
) -> unique_inverse_result[_DTypeT]: ...
def unique_values(x: array[Any, _DTypeT], /) -> array[tuple[int], _DTypeT]: ...

_DTypeT = TypeVar("_DTypeT", bound=Dtype)
_DTypeT_co = TypeVar("_DTypeT_co", bound=Dtype, covariant=True)



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import TypeVar

import numpy as np

from ._spec_array_object import array
from ._typing import Dtype, Shape

def argsort(
    x: array[_ShapeT],
    /,
    *,
    axis: int = ...,
    descending: bool = ...,
    stable: bool = ...,
) -> array[_ShapeT, Dtype[np.int64]]: ...
def sort(
    x: _ArrayT,
    /,
    *,
    axis: int = ...,
    descending: bool = ...,
    stable: bool = ...,
) -> _ArrayT: ...

_ArrayT = TypeVar("_ArrayT", bound=array)
_ShapeT = TypeVar("_ShapeT", bound=Shape)



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__all__ = [
    "from_cf_contiguous",
    "from_cf_indexed",
    "to_cf_contiguous",
    "to_cf_indexed",
]

from typing import Any, TypeVar

import numpy as np
from ragged._spec_array_object import array
from ragged._typing import Dtype  # noqa: PLC2701

def to_cf_contiguous(
    x: array[tuple[int, int | None], _DTypeT],
) -> tuple[array[tuple[int], _DTypeT], array[tuple[int], Dtype[np.int64]]]: ...
def from_cf_contiguous(
    content: array[tuple[int], _DTypeT],
    counts: array[tuple[int], Dtype[np.integer[Any]]],
) -> array[tuple[int, None], _DTypeT]: ...
def to_cf_indexed(
    x: array[tuple[int, int | None], _DTypeT],
) -> tuple[array[tuple[int], _DTypeT], array[tuple[int], Dtype[np.int64]]]: ...
def from_cf_indexed(
    content: array[tuple[int], _DTypeT],
    counts: array[tuple[int], Dtype[np.integer[Any]]],
) -> array[tuple[int, None], _DTypeT]: ...

_DTypeT = TypeVar("_DTypeT", bound=Dtype)



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Literal, TypeAlias, TypeVar, overload

import numpy as np
from typing_extensions import Unpack

from ._spec_array_object import array
from ._typing import Dtype

@overload
def all(  # noqa: A001
    x: array,
    /,
    *,
    axis: int | tuple[int, ...] | None = ...,
    keepdims: Literal[False] = ...,
) -> array[Any, Dtype[np.bool_]]: ...
@overload
def all(  # noqa: A001
    x: array[_ShapeT],
    /,
    *,
    axis: int | tuple[int, ...] | None = ...,
    keepdims: Literal[True],
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def all(  # noqa: A001
    x: array,
    /,
    *,
    axis: int | tuple[int, ...] | None = ...,
    keepdims: bool,
) -> array[Any, Dtype[np.bool_]]: ...

@overload
def any(  # noqa: A001
    x: array,
    /,
    *,
    axis: int | tuple[int, ...] | None = ...,
    keepdims: Literal[False] = ...,
) -> array[Any, Dtype[np.bool_]]: ...
@overload
def any(  # noqa: A001
    x: array[_ShapeT],
    /,
    *,
    axis: int | tuple[int, ...] | None = ...,
    keepdims: Literal[True],
) -> array[_ShapeT, Dtype[np.bool_]]: ...
@overload
def any(  # noqa: A001
    x: array,
    /,
    *,
    axis: int | tuple[int, ...] | None = ...,
    keepdims: bool,
) -> array[Any, Dtype[np.bool_]]: ...

_Axis: TypeAlias = int | None
_ShapeT = TypeVar(
    "_ShapeT",
    tuple[()],
    tuple[int],
    tuple[int, int],
    tuple[int, int, int],
    tuple[int, int, int, int],
    tuple[int, int, int, int, Unpack[tuple[int, ...]]],
    tuple[int, int, int, Unpack[tuple[int, ...]]],
    tuple[int, int, Unpack[tuple[int, ...]]],
    tuple[int, Unpack[tuple[int, ...]]],
    tuple[int, ...],
    tuple[int, _Axis],
    tuple[int, _Axis, _Axis],
    tuple[int, _Axis, _Axis, _Axis],
    tuple[int, _Axis, _Axis, _Axis, Unpack[tuple[_Axis, ...]]],
    tuple[int, _Axis, _Axis, Unpack[tuple[_Axis, ...]]],
    tuple[int, _Axis, Unpack[tuple[_Axis, ...]]],
    tuple[int, Unpack[tuple[_Axis, ...]]],
)



# Copyright 2025 hingebase

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Any, TypeAlias, TypeVar, overload

import numpy as np
import numpy.typing as npt
from typing_extensions import TypeAliasType

from ._spec_array_object import array
from ._typing import Dtype, Shape

def abs(x: _ArrayT, /) -> _ArrayT: ...  # noqa: A001

@overload
def acos(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def acos(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def acos(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def acos(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def acosh(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def acosh(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def acosh(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def acosh(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

def add(x1: array, x2: array, /) -> array: ...

@overload
def asin(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def asin(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def asin(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def asin(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def asinh(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def asinh(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def asinh(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def asinh(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def atan(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def atan(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def atan(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def atan(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

def atan2(x1: _RealArray, x2: _RealArray, /) -> _NDArray[np.floating[Any]]: ...

@overload
def atanh(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def atanh(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def atanh(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def atanh(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

def bitwise_and(
    x1: _BoolOrIntArray,
    x2: _BoolOrIntArray,
    /,
) -> _BoolOrIntArray: ...
def bitwise_invert(x: _BoolOrIntArrayT, /) -> _BoolOrIntArrayT: ...
def bitwise_left_shift(
    x1: _BoolOrIntArray,
    x2: _BoolOrIntArray,
    /,
) -> _NDArray[np.integer[Any]]: ...
def bitwise_or(
    x1: _BoolOrIntArray,
    x2: _BoolOrIntArray,
    /,
) -> _BoolOrIntArray: ...
def bitwise_right_shift(
    x1: _BoolOrIntArray,
    x2: _BoolOrIntArray,
    /,
) -> _NDArray[np.integer[Any]]: ...
def bitwise_xor(
    x1: _BoolOrIntArray,
    x2: _BoolOrIntArray,
    /,
) -> _BoolOrIntArray: ...

def ceil(x: _RealArrayT, /) -> _RealArrayT: ...
def conj(x: _NumericArrayT, /) -> _NumericArrayT: ...

@overload
def cos(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def cos(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def cos(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def cos(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def cosh(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def cosh(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def cosh(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def cosh(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

def divide(x1: array, x2: array, /) -> _InexactArray: ...
def equal(x1: array, x2: array, /) -> _BoolArray: ...

@overload
def exp(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def exp(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def exp(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def exp(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def expm1(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def expm1(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def expm1(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def expm1(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

def floor(x: _RealArrayT, /) -> _RealArrayT: ...
def floor_divide(
    x1: _RealArray,
    x2: _RealArray,
    /,
) -> _NDArray[np.integer[Any] | np.floating[Any]]: ...
def greater(x1: array, x2: array, /) -> _BoolArray: ...
def greater_equal(x1: array, x2: array, /) -> _BoolArray: ...

@overload
def imag(
    x: array[_ShapeT, Dtype[np.complexfloating[_BitT, _BitT]]],
    /,
) -> array[_ShapeT, Dtype[np.floating[_BitT]]]: ...
@overload
def imag(
    x: array[_ShapeT, Dtype[np.int32 | np.uint32 | np.float32]],
    /,
) -> _Float16Array[_ShapeT]: ...
@overload
def imag(
    x: array[_ShapeT, Dtype[np.int64 | np.uint64 | np.float64]],
    /,
) -> _Float32Array[_ShapeT]: ...

def isfinite(x: array[_ShapeT]) -> array[_ShapeT, Dtype[np.bool_]]: ...
def isinf(x: array[_ShapeT]) -> array[_ShapeT, Dtype[np.bool_]]: ...
def isnan(x: array[_ShapeT]) -> array[_ShapeT, Dtype[np.bool_]]: ...
def less(x1: array, x2: array, /) -> _BoolArray: ...
def less_equal(x1: array, x2: array, /) -> _BoolArray: ...

@overload
def log(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def log(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def log(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def log(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def log1p(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def log1p(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def log1p(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def log1p(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def log2(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def log2(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def log2(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def log2(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def log10(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def log10(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def log10(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def log10(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

def logaddexp(x1: _RealArray, x2: _RealArray, /) -> _NDArray[np.floating[Any]]:
    ...
def logical_and(x1: array, x2: array, /) -> _BoolArray: ...
def logical_not(x: array[_ShapeT]) -> array[_ShapeT, Dtype[np.bool_]]: ...
def logical_or(x1: array, x2: array, /) -> _BoolArray: ...
def logical_xor(x1: array, x2: array, /) -> _BoolArray: ...
def multiply(x1: array, x2: array, /) -> array: ...
def negative(x: _NumericArrayT, /) -> _NumericArrayT: ...
def not_equal(x1: array, x2: array, /) -> _BoolArray: ...
def positive(x: _NumericArrayT, /) -> _NumericArrayT: ...
def pow(x1: array, x2: array, /) -> _NumericArray: ...  # noqa: A001

@overload
def real(
    x: array[_ShapeT, Dtype[np.complexfloating[_BitT, _BitT]]],
    /,
) -> array[_ShapeT, Dtype[np.floating[_BitT]]]: ...
@overload
def real(
    x: array[_ShapeT, Dtype[np.int32 | np.uint32 | np.float32]],
    /,
) -> _Float16Array[_ShapeT]: ...
@overload
def real(
    x: array[_ShapeT, Dtype[np.int64 | np.uint64 | np.float64]],
    /,
) -> _Float32Array[_ShapeT]: ...

def remainder(
    x1: _RealArray,
    x2: _RealArray,
    /,
) -> _NDArray[np.integer[Any] | np.floating[Any]]: ...

@overload
def round(x: _InexactArrayT, /) -> _InexactArrayT: ...  # noqa: A001
@overload
def round(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...  # noqa: A001
@overload
def round(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...  # noqa: A001
@overload
def round(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...  # noqa: A001

def sign(x: _NumericArrayT, /) -> _NumericArrayT: ...

@overload
def sin(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def sin(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def sin(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def sin(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def sinh(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def sinh(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def sinh(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def sinh(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def square(x: _NumericArrayT, /) -> _NumericArrayT: ...
@overload
def square(
    x: array[_ShapeT, Dtype[np.bool_]],
    /,
) -> array[_ShapeT, Dtype[np.int8]]: ...

@overload
def sqrt(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def sqrt(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def sqrt(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def sqrt(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def subtract(x1: _NumericArray, x2: array, /) -> _NumericArray: ...
@overload
def subtract(x1: array, x2: _NumericArray, /) -> _NumericArray: ...

@overload
def tan(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def tan(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def tan(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def tan(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

@overload
def tanh(x: _InexactArrayT, /) -> _InexactArrayT: ...
@overload
def tanh(x: _ToFloat16Array[_ShapeT], /) -> _Float16Array[_ShapeT]: ...
@overload
def tanh(x: _ToFloat32Array[_ShapeT], /) -> _Float32Array[_ShapeT]: ...
@overload
def tanh(x: _ToFloat64Array[_ShapeT], /) -> _Float64Array[_ShapeT]: ...

def trunc(x: _RealArrayT, /) -> _RealArrayT: ...

_ArrayT = TypeVar("_ArrayT", bound=array)
_BitT = TypeVar("_BitT", bound=npt.NBitBase)
_SCT = TypeVar("_SCT", bound=np.bool_ | np.number[Any])
_NDArray = TypeAliasType(
    "_NDArray",
    array[Any, Dtype[_SCT]],
    type_params=(_SCT,),
)
_BoolArray: TypeAlias = _NDArray[np.bool_]
_BoolOrIntArray: TypeAlias = _NDArray[np.bool_ | np.integer[Any]]
_BoolOrIntArrayT = TypeVar("_BoolOrIntArrayT", bound=_BoolOrIntArray)
_ShapeT = TypeVar("_ShapeT", bound=Shape)
_Float16Array: TypeAlias = array[_ShapeT, Dtype[np.float16]]
_Float32Array: TypeAlias = array[_ShapeT, Dtype[np.float32]]
_Float64Array: TypeAlias = array[_ShapeT, Dtype[np.float64]]
_InexactArray: TypeAlias = _NDArray[np.inexact[Any]]
_InexactArrayT = TypeVar("_InexactArrayT", bound=_InexactArray)
_NumericArray: TypeAlias = _NDArray[np.number[Any]]
_NumericArrayT = TypeVar("_NumericArrayT", bound=_NumericArray)
_RealArray: TypeAlias = _NDArray[np.bool_ | np.integer[Any] | np.floating[Any]]
_RealArrayT = TypeVar("_RealArrayT", bound=_RealArray)
_ToFloat16Array: TypeAlias = array[
    _ShapeT,
    Dtype[np.int8 | np.uint8 | np.bool_],
]
_ToFloat32Array: TypeAlias = array[_ShapeT, Dtype[np.int16 | np.uint16]]
_ToFloat64Array: TypeAlias = array[
    _ShapeT,
    Dtype[np.int32 | np.uint32 | np.int64 | np.uint64],
]
